import java.io.File;
import java.util.ArrayList;
import java.util.List;

import javafx.animation.FadeTransition;
import javafx.animation.Interpolator;
import javafx.animation.KeyFrame;
import javafx.animation.KeyValue;
import javafx.animation.Timeline;
import javafx.animation.TimelineBuilder;
import javafx.application.Application;
import javafx.beans.binding.Bindings;
import javafx.beans.binding.IntegerBinding;
import javafx.beans.property.DoubleProperty;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.SimpleDoubleProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.Group;
import javafx.scene.GroupBuilder;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.SceneBuilder;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.effect.DropShadow;
import javafx.scene.effect.GaussianBlur;
import javafx.scene.effect.Reflection;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.input.MouseEvent;
import javafx.scene.media.AudioClip;
import javafx.scene.media.Media;
import javafx.scene.media.MediaPlayer;
import javafx.scene.paint.Color;
import javafx.scene.paint.CycleMethod;
import javafx.scene.paint.LinearGradient;
import javafx.scene.paint.Stop;
import javafx.scene.shape.Circle;
import javafx.scene.shape.Path;
import javafx.scene.shape.Rectangle;
import javafx.scene.shape.Shape;
import javafx.scene.text.Font;
import javafx.stage.Stage;
import javafx.util.Duration;

public class Hunter extends Application {
	private static final File MEDIA_DIR = new File(
			System.getProperty("user.dir"));
	private static final int LEVEL_INCREASE = 5;
	private static final int START_LEVEL = 1;
	private static final int INITIAL_SCORE = 0;
	private static final int INITIAL_LIVES = 3;
	private final IntegerProperty score = new SimpleIntegerProperty(
			INITIAL_SCORE);
	private final IntegerProperty lives = new SimpleIntegerProperty(
			INITIAL_LIVES);
	private final IntegerBinding level = score.divide(LEVEL_INCREASE).add(
			START_LEVEL);
	private final List<Timeline> moveTimelines = new ArrayList<>();
	private Group enemyGroup;
	private Scene scene;
	private Timeline createEnemyTimeline;
	private Group particleGroup;
	private MediaPlayer backgroundSound;
	private AudioClip explodeClip;
	private AudioClip escapeClip;

	public static void main(String[] args) {
		Application.launch(args);
	}

	@Override
	public void start(final Stage primaryStage) throws Exception {
		primaryStage.setTitle("Hunter");

		level.addListener(new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable,
					Number oldValue, Number newValue) {
				createEnemyTimeline.stop();
				startCreatingEnemies();
			}
		});

		final DoubleProperty mouseX = new SimpleDoubleProperty();

		Rectangle background = new Rectangle();
		background.setOnMouseMoved(new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent event) {
				mouseX.set(event.getSceneX());
			}
		});

		enemyGroup = new Group();

		enemyGroup.effectProperty().bind(
				Bindings.when(lives.isEqualTo(0)).then(new GaussianBlur())
						.otherwise((GaussianBlur) null));

		particleGroup = new Group();

		Label information = new Label();
		information.setFont(new Font(24));
		information.setTextFill(Color.WHITESMOKE);
		information.textProperty().bind(
				Bindings.format("Score: %s, Level %s, Lives: %s", score, level,
						lives));

		Button restartButton = new Button("Restart");

		// final Rectangle ship = new Rectangle(80, 10);

		final ImageView ship = new ImageView(new Image(getClass()
				.getResourceAsStream("sswlogo.gif")));

		ship.xProperty().bind(mouseX.subtract(ship.getImage().widthProperty().divide(2)));
//		ship.setFill(Color.BLUE);

		final Group shotsGroup = new Group();

		scene = SceneBuilder
				.create()
				.width(600)
				.height(400)
				.root(GroupBuilder
						.create()
						.children(background, ship, information, shotsGroup,
								enemyGroup, particleGroup, restartButton)
						.build()).build();

		background.widthProperty().bind(scene.widthProperty());
		background.heightProperty().bind(scene.heightProperty());

		background.setOnMousePressed(new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent event) {
				if (event.isPrimaryButtonDown()) {
					Circle shot = new Circle(event.getSceneX(), ship.getY(), 3,
							Color.GREEN);
					shotsGroup.getChildren().add(shot);
				} else {
					primaryStage.setFullScreen(true);
				}
			}
		});

		ship.yProperty().bind(
				scene.heightProperty().subtract(ship.getImage().heightProperty()));

		restartButton.translateXProperty().bind(
				scene.widthProperty().subtract(restartButton.widthProperty())
						.divide(2));

		restartButton.translateYProperty().bind(
				scene.heightProperty().subtract(restartButton.heightProperty())
						.divide(2));

		restartButton.visibleProperty().bind(lives.isEqualTo(0));

		restartButton.setOnAction(new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				restartGame();
			}
		});

		loadSounds();

		lives.addListener(new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable,
					Number oldValue, Number newValue) {
				if (newValue.intValue() == 0) {
					stopGame();
				}
			}
		});

		backgroundSound.play();

		Timeline updateShotsTimeline = new Timeline();
		updateShotsTimeline.setCycleCount(Timeline.INDEFINITE);
		updateShotsTimeline.getKeyFrames().add(
				new KeyFrame(Duration.millis(10),
						new EventHandler<ActionEvent>() {
							@Override
							public void handle(ActionEvent event) {
								List<Node> shotsToRemove = new ArrayList<>();
								for (Node n : shotsGroup.getChildren()) {
									Circle shot = (Circle) n;
									shot.setCenterY(shot.getCenterY() - 3);
									if (shot.getCenterY() < 0) {
										shotsToRemove.add(shot);
									}
									List<Shape> enemiesToKill = new ArrayList<>();
									for (Node e : enemyGroup.getChildren()) {
										Shape ufo = (Shape) e;
										if (ufo.getBoundsInParent().contains(
												shot.getCenterX(),
												shot.getCenterY())) {
											enemiesToKill.add(ufo);
										}
									}
									for (Shape e : enemiesToKill) {
										kill(e, shot.getCenterX(),
												shot.getCenterY());
									}
								}
								shotsGroup.getChildren().removeAll(
										shotsToRemove);
							}
						}));

		updateShotsTimeline.play();

		startCreatingEnemies();

		primaryStage.setScene(scene);

		primaryStage.show();
	}

	private void loadSounds() {
		backgroundSound = new MediaPlayer(new Media(new File(MEDIA_DIR,
				"32157___zin___distriff_2.mp3").toURI().toString()));
		backgroundSound.setCycleCount(MediaPlayer.INDEFINITE);

		escapeClip = new AudioClip(new File(MEDIA_DIR,
				"3060__SpeedY__UFOblip.mp3").toURI().toString());

		explodeClip = new AudioClip(new File(MEDIA_DIR,
				"35458__jobro__Explosion_1.mp3").toURI().toString());
	}

	private void startCreatingEnemies() {
		createEnemyTimeline = TimelineBuilder
				.create()
				.cycleCount(Timeline.INDEFINITE)
				.keyFrames(
						new KeyFrame(
								Duration.millis(100 + 1000d / level.get()),
								new EventHandler<ActionEvent>() {
									@Override
									public void handle(ActionEvent event) {
										createEnemy();
									}
								})).build();

		createEnemyTimeline.play();
	}

	private void createEnemy() {
		final Circle mainCircle = new Circle(40);
		final Shape subtractCircles = Path.union(new Circle(-20, 20, 40),
				new Circle(20, 20, 40));

		final Shape ufo = Path.subtract(mainCircle, subtractCircles);

		ufo.setTranslateY(Math.random() * scene.getHeight());

		ufo.setLayoutY(-ufo.getBoundsInLocal().getMinY()
				- ufo.getBoundsInLocal().getHeight() / 2);

		// final Circle circle = new Circle(0, Math.random() *
		// scene.getHeight(),
		// 20);
		//
		ufo.setFill(new LinearGradient(0, 0, 1, 1, true, CycleMethod.NO_CYCLE,
				new Stop(0, Color.BLACK), new Stop(1, Color.CRIMSON)));
		ufo.setStroke(Color.BLUEVIOLET);

		DropShadow shadow = new DropShadow();
		shadow.setOffsetX(-5);
		shadow.setOffsetY(5);
		shadow.setColor(Color.YELLOWGREEN);

		Reflection reflection = new Reflection();
		reflection.setInput(shadow);

		ufo.setEffect(reflection);

		ufo.setOnMousePressed(new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent event) {
				if (lives.get() > 0) {
					kill(ufo, event.getSceneX(), event.getSceneY());
				}
			}
		});

		enemyGroup.getChildren().add(ufo);

		Timeline move = new Timeline();
		move.getKeyFrames().add(
				new KeyFrame(Duration.seconds(3),
						new EventHandler<ActionEvent>() {
							@Override
							public void handle(ActionEvent event) {
								if (enemyGroup.getChildren().remove(ufo)) {
									lives.set(lives.get() - 1);
									escapeClip.play();
								}
							}
						}, new KeyValue(ufo.translateXProperty(), scene
								.getWidth(), Interpolator.EASE_BOTH)));
		move.play();

		moveTimelines.add(move);

	}

	private void kill(final Shape ufo, double x, double y) {
		enemyGroup.getChildren().remove(ufo);
		score.set(score.get() + 1);
		explodeClip.play();
		showExplosion(x, y);
	}

	private void showExplosion(double x, double y) {
		DoubleProperty dist = new SimpleDoubleProperty();

		final Group particles = new Group();

		for (int i = 0; i < 20; ++i) {
			Circle p = new Circle(3);
			p.setFill(Color.RED);
			double speedX = Math.random() - 0.5;
			double speedY = Math.random() - 0.5;
			p.centerXProperty().bind(dist.multiply(speedX).add(x));
			p.centerYProperty().bind(dist.multiply(speedY).add(y));
			particles.getChildren().add(p);
		}

		Timeline explode = new Timeline();
		explode.getKeyFrames().add(
				new KeyFrame(Duration.millis(500),
						new EventHandler<ActionEvent>() {
							@Override
							public void handle(ActionEvent event) {
								particleGroup.getChildren().remove(particles);
							}
						}, new KeyValue(dist, 200)));

		explode.play();

		FadeTransition fade = new FadeTransition();
		fade.setDuration(Duration.millis(500));
		fade.setNode(particles);
		fade.setFromValue(1);
		fade.setToValue(0);

		fade.play();

		particleGroup.getChildren().add(particles);

	}

	private void restartGame() {
		enemyGroup.getChildren().clear();
		lives.set(INITIAL_LIVES);
		score.set(INITIAL_SCORE);
		createEnemyTimeline.play();
	}

	private void stopGame() {
		createEnemyTimeline.stop();
		for (Timeline move : moveTimelines) {
			move.pause();
		}
		moveTimelines.clear();
	}
}
